/** **********************************************************************
 * ConsoleDisplay.java
 * A basic question-and-answer dialog GUI
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;

import java.awt.*;
import java.applet.Applet;
import java.util.*;
import java.io.*;

// //////////////////////////////////////////////////////////////////////
// This class includes a Panel containing input and output text areas that can
// be plugged into Jess as a ReteDisplay. It uses the TextInputStream and
// TextAreaOutputStream classes to turn textual data into I/O streams.
// //////////////////////////////////////////////////////////////////////

public class ConsoleDisplay extends Observable implements ReteDisplay
{
  // The Panel
  private Panel m_panel;

  // Members used for presenting output
  private TextArea m_ta;
  private PrintStream m_out;
  private Button m_bClear;

  // Members used for getting input
  private TextField m_tf;

  // TextInputStream is in the jess package
  private TextInputStream m_in;

  // Null if we're in an application.
  private Applet m_applet;
  
  // ************************************************************
  // Lay out the Panel in the constructor
  // OK to pass 'null' for argument!
  // ************************************************************

  public ConsoleDisplay(Applet theApplet)
  {
    m_applet = theApplet;

    // Set up the GUI elements
    m_ta = new TextArea(10, 40);
    m_tf = new TextField(50);
    m_ta.setEditable(false);
    m_bClear = new Button("Clear Window");
    Panel p = new Panel();
    p.setLayout(new BorderLayout());

    // Set up I/O streams
    m_out = new PrintStream(new TextAreaOutputStream(m_ta), true);
    m_in = new TextInputStream();
    
    // Assemble the GUI
    m_panel = new ConsolePanel(m_ta, m_tf, m_out, m_in, m_bClear);
    m_panel.setLayout(new BorderLayout());
    m_panel.add("Center", m_ta);
    p.add("Center", m_tf);
    p.add("East", m_bClear);
    m_panel.add("South", p);
    
  }

  final public Panel panel() { return m_panel; }
  
  final public void setFocus() { m_tf.requestFocus(); }

  // ************************************************************
  // These are the actual ReteDisplay methods
  // Here we hand out our specialized streams
  // ************************************************************

  public PrintStream stdout() { return m_out;} 
  public PrintStream stderr() { return m_out;} 
  public InputStream stdin()  { return m_in; }

  public Applet applet() { return m_applet; }

  // ************************************************************
  // All this notification stuff is for the benefit of the 'View' command.
  // ************************************************************

  public void assertFact(ValueVector fact)
  {
    setChanged();
    notifyObservers("FACT");
  }
  public void retractFact(ValueVector fact)
  {
    setChanged();
    notifyObservers("FACT");
  }

  public void addDeffacts(Deffacts df) {}
  public void addDeftemplate(Deftemplate dt) {}

  public void addDefrule(Defrule rule)
  {
    setChanged();
    notifyObservers("RULE");
  }
  public void activateRule(Defrule rule) {}
  public void deactivateRule(Defrule rule) {}
  public void fireRule(Defrule rule)
  {
    setChanged();
    notifyObservers("RULE");
    System.gc();
  }
}

// **********************************************************************
// When we go to Java 1.1, this class can go away and we can make
// ConsoleDisplay an ActionListener
// **********************************************************************

class ConsolePanel extends Panel
{
  private TextArea m_ta;
  private TextField m_tf;
  private PrintStream m_out;
  private TextInputStream m_in;
  private Button m_bClear;

  ConsolePanel(TextArea ta, TextField tf, PrintStream out,
               TextInputStream in, Button bClear) 
  { 
    m_ta = ta; m_tf = tf; m_out = out; m_in = in; m_bClear = bClear;
  }

  // ************************************************************
  // When the user hits RETURN in the text field, attach the text
  // to the stream and clear the field.
  // ************************************************************

  public boolean action(Event e, Object o)
  {
    if (e.target == m_tf)
      {
        synchronized (m_ta)
          {  
            m_out.println(m_tf.getText());
          }

        m_in.appendText(m_tf.getText() + "\n");
        m_tf.setText("");    
        return true;
      }
    else if (e.target == m_bClear)
      {
        m_ta.setText("");
        m_tf.setText("");
        return true;
      }
    return false;
  }


}
