/** **********************************************************************
 * A class for parsing, assembling, and interpreting function calls
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;

import java.io.*;
import java.net.*;
import java.applet.*;
import java.util.*;

public class Funcall extends ValueVector
{

  public String toString() { return "(" + super.toString() + ")"; }

  static Value s_true;
  static Value s_false;
  static Value s_nil;
  static Value s_nillist;
  static Value s_else;
  static Value s_then;
  static Value s_do;
  static Value s_eof;

  static
  {
    try
      {
        s_true = new Value("TRUE",RU.ATOM);
        s_false = new Value("FALSE",RU.ATOM);
        s_nil = new Value("nil",RU.ATOM);
        s_nillist = new Value(new ValueVector(), RU.LIST);
        s_else = new Value("else",RU.ATOM);
        s_then = new Value("then",RU.ATOM);
        s_do = new Value("do",RU.ATOM);
        s_eof = new Value("EOF",RU.ATOM);
      }
    catch (ReteException re)
      {
        System.out.println("*** FATAL ***: Can't instantiate constants");
        System.exit(0);
      }

  }
  
  // Load in all the intrinsic functions

  static void loadIntrinsics(Rete engine) throws ReteException
  {
    String [] intlist =
    {"Assert","Retract", "RetractString", 
     "Printout", "ExtractGlobal",
     "Open", "Close", "Foreach",
     "Read","Readline","GensymStar","While",
     "If","Bind","Modify","And","Or","Not",
     "Eq", "EqStar", "Equals","NotEquals","Gt","Lt",
     "GtOrEq","LtOrEq","Neq","Mod","Plus",
     "Times","Minus","Divide","SymCat",
     "LoadFacts","SaveFacts",
     "AssertString","UnDefrule", "Try"};
    
    int i = 0;

    try
      {
        for (i=0; i< intlist.length; i++)
          {
            Userfunction uf = (Userfunction)
              Class.forName("jess." + intlist[i]).newInstance();
            engine.addUserfunction(uf);
          }

        engine.addUserfunction(new JessVersion(JessVersion.NUMBER));
        engine.addUserfunction(new JessVersion(JessVersion.STRING));

        engine.addUserfunction(new HaltEtc(HaltEtc.HALT));
        engine.addUserfunction(new HaltEtc(HaltEtc.EXIT));
        engine.addUserfunction(new HaltEtc(HaltEtc.CLEAR));
        engine.addUserfunction(new HaltEtc(HaltEtc.RUN));
        engine.addUserfunction(new HaltEtc(HaltEtc.RESET));        
        engine.addUserfunction(new HaltEtc(HaltEtc.RETURN));        

        engine.addUserfunction(new Watch(Watch.WATCH));
        engine.addUserfunction(new Watch(Watch.UNWATCH));

        engine.addUserfunction(new FactsRules(FactsRules.FACTS));
        engine.addUserfunction(new FactsRules(FactsRules.RULES));

        engine.addUserfunction(new StoreFetch(StoreFetch.STORE));
        engine.addUserfunction(new StoreFetch(StoreFetch.FETCH));

      }
    catch (Throwable t)
      {
        throw new ReteException("Funcall::loadIntrisics",
                                "Missing intrinsic function class:",
                                intlist[i]);
      }
  }

  Userfunction m_funcall;

  public Funcall(String name, Rete engine) throws ReteException
  {
    add(new Value(name, RU.ATOM));

    m_funcall = engine.findUserfunction(name);
  }
    

  Funcall(int size)
  {
    super(size);
  }
  
  public Object clone()
  {
    Funcall vv = new Funcall(m_ptr);
    vv.m_ptr = m_ptr;
    System.arraycopy(m_v, 0, vv.m_v, 0, m_ptr);
    vv.m_funcall = m_funcall;
    return vv;
  }

  public Funcall cloneInto(Funcall vv)
  {
    if (m_ptr > vv.m_v.length) 
      {
        Value[] nv = new Value[m_ptr];
        vv.m_v = nv;
      }
    vv.m_ptr = m_ptr;
    System.arraycopy(m_v, 0, vv.m_v, 0, m_ptr);
    vv.m_funcall = m_funcall;
    return vv;
  }

  public final static Value TRUE()
  {
    return s_true;
  }
  
  public final static Value FALSE()
  {
    return s_false;
  }
  
  public final static Value NIL()
  {
    return s_nil;
  }

  public final static Value NILLIST()
  {
    return s_nillist;
  }

  public final static Value EOF()
  {
    return s_eof;
  }
  
  
  /**
    Add an argument to this funcall
    */

  void addArgument(String value, int type) throws ReteException
  {
    add(new Value(value, type));
  }

  void addArgument(int value, int type) throws ReteException
  {
    add(new Value(value, type));
  }

  void addArgument(double value, int type) throws ReteException 
  {
    add(new Value(value, type));
  }

  void addArgument(Funcall value) throws ReteException
  {
    add(new Value(value, RU.FUNCALL));
  }

  void addArgument(Fact value) throws ReteException
  {
    ValueVector vv = value.factData();
    add(new Value(vv, vv.get(RU.DESC).descriptorValue()));
  }
  
  void addArgument(ValueVector value, int type) throws ReteException
  {
    add(new Value(value, type));
  }
  
  /**
    Execute a Funcall, recursively expanding its fields in the context of a
    Defrule. 

    @returns A Value structure containing the return value, or null
    if expand_root false.
    */

  static final Value execute(Funcall vv, Context context)
       throws ReteException
  { return execute(vv, context, null, null); }

  static final Value execute(Funcall vv, Context context, EvalCache vs)
       throws ReteException
  { return execute(vv, context, vs, null); }

  static Value execute(Funcall vv, Context context, EvalCache vs, Value retval)
       throws ReteException
  {

    int omark = (vs != null ? vs.markValue() : 0);

    Userfunction uf = vv.m_funcall;
    if (uf == null)
      {
        String name = vv.get(0).stringValue();
        
        if ((uf = vv.m_funcall = context.m_engine.findUserfunction(name))
            == null)
          throw new ReteException("Funcall::simpleExecute",
                                  "Unimplemented function",
                                  name);
      }
    
    int firstUnexpandedArg;

    if (uf instanceof ControlStructure)
      {
        context.clearReturnValue();
        firstUnexpandedArg = ((ControlStructure) uf).skipFromHere();
      }
    else
      firstUnexpandedArg = vv.size();

    // for each argument
    for (int i=1; i<firstUnexpandedArg; i++)
      {
        int type = vv.get(i).type();
        switch(type)
          {
          case RU.FUNCALL:
            {
              
              // dig out the subexpression
              Funcall subexp = vv.get(i).funcallValue();
              
              // expand subexpression, keeping return value
              Value rv = (vs != null) ? vs.getValue() :
                new Value(0, RU.INTEGER);
              int mark = (vs != null) ? vs.markValue() : 0;
              Value r = execute(subexp, context, vs, rv);
              vv.set(r,i);
              if (vs != null)
                vs.restoreValue(mark);
              
              break;
            }
          case RU.UNORDERED_FACT:
          case RU.ORDERED_FACT:
            {
              
              ValueVector fact = vv.get(i).factValue();
              for (int j=RU.FIRST_SLOT; j<fact.size(); j++)
                {
                  if (fact.get(j).type() == RU.FUNCALL)
                    {
                      Funcall subexp = fact.get(j).funcallValue();
                      Value rv = (vs != null) ? vs.getValue() :
                        new Value(0, RU.INTEGER);
                      int mark = (vs != null) ? vs.markValue() : 0;
                      Value r = execute(subexp, context, vs, rv);
                      fact.set(r,j);
                      if (vs != null)
                        vs.restoreValue(mark);
                    }
                }
              break;
            }
          case RU.LIST:
            {
              
              ValueVector list = vv.get(i).listValue();
              for (int j=1; j<list.size(); j++)
                {
                  if (list.get(j).type() == RU.FUNCALL)
                    {
                      Value rv = (vs != null) ? vs.getValue() :
                        new Value(0, RU.INTEGER);
                      int mark = (vs != null) ? vs.markValue() : 0;
                      Funcall subexp = list.get(j).funcallValue();
                      Value r = execute(subexp, context, vs, rv);
                      list.set(r,j);
                      if (vs != null)
                        vs.restoreValue(mark);
                    }
                }
              vv.set(new Value(Context.flattenList(list), RU.LIST), i);
              break;
            }
          }
      }
    // we need to expand ourselves too
    Value v = simpleExecute(vv, context, retval);
    if (vs != null)
      vs.restoreValue(omark);
    return v;
  }
  
  /**
    Execute a funcall in a particular context
    We let the Value class do all our argument type checking, leading
    to remarkably small code!

    Very simple version is the only public one.
    */
  
  public static Value simpleExecute(Funcall vv, Context context)
       throws ReteException
  { return simpleExecute(vv, context, null); }

  static Value simpleExecute(Funcall vv, Context context, Value v)
       throws ReteException
  {
    Userfunction uf = vv.m_funcall;
    if (uf == null)
      {
        Rete engine = context.m_engine;
        String name = vv.get(0).stringValue();
        
        if ((uf = vv.m_funcall = engine.findUserfunction(name)) == null)
          throw new ReteException("Funcall::simpleExecute",
                                  "Unimplemented function",
                                  name);
      }
    
    if (uf instanceof Fastfunction && v != null)
      {
        return ((Fastfunction) uf).call(vv, context, v);
      }
    else
      {
        return uf.call(vv, context);
      }
    
  }
  
  /* **********************************************************************
   * Code to pretty print funcalls
   * Original by Rajaram Ganeshan:
   * 1. public String ppFuncall()
   * 2. private String ppArgument ( ValueVector vv, int i)
   ********************************************************************** */

  String ppFuncall(Rete engine) throws ReteException
  {
    // functor
    StringBuffer sb = new StringBuffer(100);
    sb.append("(");
    sb.append(get(0).stringValue());
    sb.append(" ");

    // for each argument
    for (int i=1; i< size(); i++)
      {
        sb.append(ppArgument(i, engine));
        sb.append(" ");
      }
    sb.append(")");
    return sb.toString();
  }

  private String ppArgument (int i, Rete engine) throws ReteException
  {
    ValueVector f;
    Fact fact;
    int type = get(i).type();
    switch (type) {
    case RU.ATOM:
    case RU.FLOAT:
    case RU.INTEGER:
    case RU.VARIABLE:
      return get(i).toString();

    case RU.STRING:      
      return "\"" + get(i).stringValue() + "\"";

    case RU.LIST:
      {
        ValueVector list = get(i).listValue();
        String s = "";
          for (int j=1; j<list.size(); j++)
            s = s + ppArgument(j, engine) + " ";
        return s;
      }

    case RU.FUNCALL:
      Funcall fc = get(i).funcallValue();
      return fc.ppFuncall(engine);

    case RU.ORDERED_FACT:
    case RU.UNORDERED_FACT:
      f = get(i).factValue();
      fact = new Fact(f, engine);
      return fact.toString();

    default:
      return get(i).toString();
    }
  }

}    

// *** Fastfunction  *******************************************
abstract class Fastfunction implements Userfunction
{
  public abstract Value call(ValueVector vv, Context context, Value v)
       throws ReteException;

  public Value call(ValueVector vv, Context context) throws ReteException
  {
    return call(vv, context, new Value(0, RU.INTEGER));
  }
}

// *** assert  *******************************************
class Assert extends Fastfunction
{
  public String name() { return "assert"; }
  public Value call(ValueVector vv, Context context, Value v)
       throws ReteException
  {
    int result = -1;
    for (int i=1; i< vv.size(); i++)
      {
        ValueVector fact = vv.get(i).factValue();
        result = context.engine().assert(fact);
      }
    if (result != -1)
      return v.resetValue(result, RU.FACT_ID);
    else
      return Funcall.s_false;
  }
}
  
// *** retract *******************************************
        
class Retract implements Userfunction
{
  public String name() { return "retract"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    Value v = vv.get(1);
    if (v.type() == RU.ATOM && v.equals("*"))
      {
        context.engine().removeFacts();
      }
    else
      {
        Rete engine = context.engine();
        Userfunction uf = engine.findUserfunction("undefinstance");
        for (int i=1; i< vv.size(); i++) 
          {
            ValueVector f = engine.findFactByID(vv.get(i).factIDValue());
            if (f == null)
              continue;

            // Undefinstance, if needed.
            if (uf != null)
              {
                try
                  {
                    Fact fact = new Fact(f, engine);
                    Value ov = fact.findValue("OBJECT");
                    Funcall fc = new Funcall("undefinstance", engine);
                    fc.add(ov);
                    Funcall.simpleExecute(fc, engine.globalContext());
                  }
                catch (Exception any) { /* OK, not an object! */ }
              }
            // retract actual fact
            engine._retract(f);
          }
      }
    return Funcall.s_true;
  }
}
        
// *** printout *******************************************
      
class Printout implements Userfunction
{

Attr attr = new Attr();

  public String name() { return "printout"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String routerName = vv.get(1).stringValue();
    OutputStream os = context.engine().getOutputRouter(routerName);
    if (os == null)
      throw new ReteException("_printout::call",
                              "printout: bad router",
                              routerName);

    StringBuffer sb = new StringBuffer(100);
    int size = 1;
    String s = new String();
    for (int i = 2; i < vv.size(); i++) 
      {
        Value v = vv.get(i);
        switch(v.type()) 
          {
          case RU.ATOM:
            if (v.equals("crlf"))
              sb.append("\n");
            else {
              sb.append(v.stringValue());
            	System.out.println("data :" + v.stringValue());
            } 
	    break;
          case RU.INTEGER:
            sb.append(v.intValue());
            break;
          case RU.FLOAT:
            sb.append(v.numericValue());
            break;
          case RU.FACT_ID:
            sb.append("<Fact-");
            sb.append(v.factIDValue());
            sb.append(">");
            break;
          case RU.STRING:
            sb.append(v.stringValue());
	    if (size == 1) {
	    	s = v.stringValue();
		size ++;
		break;
	    }
	    if (size == 2) {
	    	String _s = v.stringValue();
		attr.Setting(_s, s);
		size = 1;
	    }
	    attr.Print();
            System.out.println("data1 :" + v.stringValue());
	    break;
          case RU.LIST:
            sb.append('(');
            sb.append(v.toString());
            sb.append(')');
            break;
          case RU.EXTERNAL_ADDRESS:
            sb.append(v.toString());
            break;
          case RU.VARIABLE:
          case RU.MULTIVARIABLE:
            throw new ReteException("_printout::call",
                                    "Uninitialized variable",
                                    v.toString());
          default:
            throw new ReteException("_printout::call",
                                    "Bad data type",
                                    v.toString() + " (type = "+v.type()+")");
            
          }
      }

      MadeQuery mq = new MadeQuery(attr);
      mq.Select();
    
    try
      {
        for (int i=0; i< sb.length(); i++)
          os.write(sb.charAt(i));
        os.flush();
      }
    catch (IOException ioe)
      {
        throw new ReteException("_printout::call",
                                "I/O Exception",
                                ioe.toString());
      }
    return Funcall.s_nil;
  }
}  

// *** open *******************************************
class Open implements Userfunction
{
  public String name() { return "open"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    Rete engine = context.engine();

    // Obtain parameters
    String filename = vv.get(1).stringValue();
    String router = vv.get(2).stringValue();
    String access = "r";
    if (vv.size() > 3)
      access = vv.get(3).stringValue();
    
    try
      {
        if (access.equals("r"))
          {
            engine.addInputRouter(router,
                   new BufferedInputStream(new FileInputStream(filename), 1),
                   false);
          }
        else if (access.equals("w"))
          {
            engine.addOutputRouter(router,
                   new BufferedOutputStream(new FileOutputStream(filename)));

          } 
        else if (access.equals("a"))
          {
            RandomAccessFile raf = new RandomAccessFile(filename, "rw");
            raf.seek(raf.length());
            FileOutputStream fos = new FileOutputStream(raf.getFD());
            engine.addOutputRouter(router, new BufferedOutputStream(fos));
          }
        else
          throw new ReteException("_open::call",
                                  "Unsupported access mode",
                                  access);
      }
    catch (IOException ioe)
      {
        throw new ReteException("_open::call",
                                "I/O Exception",
                                ioe.toString());
      }
    return new Value(router, RU.ATOM);
  }
}

// *** open *******************************************
class Close implements Userfunction
{
  public String name() { return "close"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    Rete engine = context.engine();
    if (vv.size() > 1)
      for (int i=1; i<vv.size(); i++)
        {
          OutputStream os;
          InputStream is;
          String router = vv.get(i).stringValue();
          try
            {
              if ((os = engine.getOutputRouter(router)) != null)
                {
                  os.close();
                  engine.removeOutputRouter(router);
                }
            }
          catch (IOException ioe) {}
          try
            {
              if ((is = engine.getInputRouter(router)) != null)
                {
                  is.close();
                  engine.removeInputRouter(router);
                }
            }
          catch (IOException ioe) {}
        }
    else
      throw new ReteException("_close::call",
                              "Must close files by name", "");
    
    return Funcall.TRUE();
  }
}

// *** read *******************************************
class Read implements Userfunction
{
  public String name() { return "read"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {

    // Find input source
    String routerName = "t";

    if (vv.size() > 1)
      routerName = vv.get(1).stringValue();

    Rete engine = context.engine();
    Tokenizer t = engine.getInputWrapper(engine.getInputRouter(routerName));

    if (t == null)
      throw new ReteException("_read::call",
                              "bad router",
                              routerName);
    JessToken jt = t.nextToken();

    // Console-like streams read a token, then throw away to newline.
    if (engine.getInputMode(routerName))
      t.discardToEOL();

    return jt.tokenToValue(null);
  }

}       
// *** readline  *******************************************
 
class Readline implements Userfunction
{
  public String name() { return "readline"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String routerName = "t";

    if (vv.size() > 1)
      routerName = vv.get(1).stringValue();

    Rete engine = context.engine();
    Tokenizer t = engine.getInputWrapper(engine.getInputRouter(routerName));

    String line = t.readLine();
    if (line == null)
      return Funcall.EOF();
    else
      return new Value(line, RU.STRING);
  }
}

// *** gensym*  *******************************************

class GensymStar implements Userfunction
{
  public String name() { return "gensym*"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    return new Value(RU.gensym("gen"), RU.ATOM);
  }
}
// *** while *******************************************

class While implements Userfunction, ControlStructure
{
  public int skipFromHere() { return 0; }
  public String name() { return "while"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    // This accepts a superset of the correct syntax...
    Value result = vv.get(1);
    if (result.type() == RU.FUNCALL)
      result = Funcall.execute(context.expandAction(result.funcallValue()),
                               context);

    // Skip optional do
    int sawDo = 0;
    if (vv.get(2).equals(Funcall.s_do))
      ++sawDo;

  outer_loop:
    while (! result.equals(Funcall.s_false)) 
      {
        for (int i=2 + sawDo; i< vv.size(); i++) 
          {
            Value current = vv.get(i);
            if (current.type() == RU.FUNCALL) 
              {
                Funcall copy = context.expandAction(current.funcallValue());
                result = Funcall.execute(copy, context);
                if (context.returning()) 
                  {
                    result = context.getReturnValue();
                    break outer_loop;
                  }
              }
            else
              result = current;
          }
        result = vv.get(1);
        if (result.type() == RU.FUNCALL)
          result = Funcall.execute(context.expandAction(result.funcallValue()),
                                   context);
      }
    return result; 
  }
}
// *** if *******************************************

class If implements Userfunction, ControlStructure
{
  public int skipFromHere() { return 2; }
  public String name() { return "if"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    // This accepts a superset of the correct syntax...

    if (!vv.get(2).equals(Funcall.s_then))
      throw new ReteException("If::call", "Expected 'then':",
                              vv.get(2).toString());
    // check condition
    Value result = vv.get(1);
    if (result.type() == RU.FUNCALL)
      result = Funcall.execute(context.expandAction(result.funcallValue()),
                               context);

    if (!(result.equals(Funcall.s_false))) 
      {
        // do 'then' part
        result = Funcall.s_false;
        for (int i=3; i< vv.size(); i++) 
          {
            Value val = vv.get(i);
            if (val.type() == RU.FUNCALL) 
              {
                Funcall copy = context.expandAction(vv.get(i).funcallValue());
                result = Funcall.execute(copy, context);
                if (context.returning()) 
                  {
                    result = context.getReturnValue();
                    break;
                  }
              }
            else
              if (val.equals(Funcall.s_else))
                break;
            else
              {
                result = val;
                break;
              }
          }
        return result; 
      }
    else
      {
        // first find the 'else'
        result = Funcall.s_false;
        boolean seen_else = false;
        for (int i=3; i< vv.size(); i++)
          {
            Value val = vv.get(i);
            if (seen_else) 
              {
                if (val.type() == RU.FUNCALL)
                  {
                    Funcall copy
                      = context.expandAction(vv.get(i).funcallValue());
                    result = Funcall.execute(copy, context);
                    if (context.returning()) 
                      {
                        result = context.getReturnValue();
                        break;
                      }
                  }
                else
                  result = val;
              } 
            else if (val.equals(Funcall.s_else))
              seen_else = true;
          }
        return result;
      }
  }
}
        
// *** bind *******************************************
        
class Bind implements Userfunction
{
  public String name() { return "bind"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    context.setVariable(vv.get(1).stringValue(),
                        vv.get(2));
    return vv.get(2);
  }
}
// *** foreach *******************************************
        
class Foreach implements Userfunction, ControlStructure
{
  public int skipFromHere() { return 3; }
  public String name() { return "foreach"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String variable = vv.get(1).stringValue();
    ValueVector items = vv.get(2).listValue();
    Value v = Funcall.NIL();

    for (int i=0; i<items.size(); i++)
      {
        context.setVariable(variable, items.get(i));
        for (int j=3; j<vv.size(); j++)
          {
            v = vv.get(j);
            switch (v.type())
              {
              case RU.FUNCALL:
                v = Funcall.execute(context.expandAction(v.funcallValue()),
                                    context);
                if (context.returning()) 
                  {
                    v = context.getReturnValue();
                    return v;
                  }                
                break;
                
              case RU.VARIABLE:
              case RU.MULTIVARIABLE:
                v = context.findBinding(v.stringValue()).m_val;
                break;
              }
          }
      }
    return v;
  }
}

// *** try *******************************************
        
class Try implements Userfunction, ControlStructure
{
  public int skipFromHere() { return 0; }
  public String name() { return "try"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    // First find catch
    int handler = -1;

    for (int j=1; j<vv.size(); j++)
      {
        if (vv.get(j).type() == RU.ATOM &&
            vv.get(j).equals("catch"))
          {
            handler = j;
            break;
          }
      }

    if (handler == -1)
      throw new ReteException("_try::call",
                              "No catch block in try expression", "");

    Value v = null;
    try
      {
        for (int j=1; j<handler; j++)
          {
            v = vv.get(j);
            switch (v.type())
              {
              case RU.FUNCALL:
                v = Funcall.execute(context.expandAction(v.funcallValue()),
                                    context);
                if (context.returning()) 
                  {
                    v = context.getReturnValue();
                    return v;
                  }                
                break;
                
              case RU.VARIABLE:
              case RU.MULTIVARIABLE:
                v = context.findBinding(v.stringValue()).m_val;            
                break;
              }
          }
        return v;
      }
    catch (Throwable t)
      {
        v = Funcall.s_false; // so we can have empty handlers
        context.setVariable("ERROR", new Value(t, RU.EXTERNAL_ADDRESS));

        for (int j=++handler; j<vv.size(); j++)
          {
            v = vv.get(j);
            switch (v.type())
              {
              case RU.FUNCALL:
                v = Funcall.execute(context.expandAction(v.funcallValue()),
                                    context);
                if (context.returning()) 
                  {
                    v = context.getReturnValue();
                    return v;
                  }                
                break;
                
              case RU.VARIABLE:
              case RU.MULTIVARIABLE:
                v = context.findBinding(v.stringValue()).m_val;            
                break;
              }
          }
        return v;
      }
  }
}

// *** modify  *******************************************
        
class Modify implements Userfunction
{
  public String name() { return "modify"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    ValueVector fact;
    if ((fact
         = context.m_engine.findFactByID(vv.get(1).factIDValue()))
        == null)
      throw new ReteException("_modify::call",
                              "modify: no such fact",
                              "");

    // we have the ValueVector fact, make a Fact out of it
        
    Fact f = new Fact(fact, context.engine());
    Rete engine = context.engine();
        
    // First find out of this is a LHS object, not an ordinary fact
    // If so, call the appropriate mutators...

    if (engine.findUserfunction("definstance") != null)
      {
        try
          {
            Value ov = f.findValue("OBJECT");
            Funcall fc = new Funcall("set", engine);
            fc.add(ov);
            fc.add(new Value("set", RU.STRING));
            fc.setLength(4);
            
            for (int i= 2; i < vv.size(); i++) 
              {
                
                // fetch the slot, value subexp, stored as a List
                ValueVector svp = vv.get(i).listValue();
                
                fc.set(svp.get(0), 2);

                if (svp.size() > 2)
                  {            
                    // wrap the new value in a list before adding
                    ValueVector mf = new ValueVector();
                    for (int j=1; j < svp.size(); j++)
                      mf.add(svp.get(j));
                    
                    fc.set(new Value(mf, RU.LIST), 3);
                  }
                else 
                  fc.set(svp.get(1), 3);
                
                Funcall.simpleExecute(fc, engine.globalContext());
              }
            return Funcall.TRUE();
          }
        catch (ReteException re) { /* OK, not an object */ }
      }
  
    // This is just an ordinary fact.
    // now change the values. For each argument...
    for (int i= 2; i < vv.size(); i++) 
      {
            
        // fetch the slot, value subexp, stored as a List
        ValueVector svp = vv.get(i).listValue();
            
        int slot_desc_idx =
          (f.findSlot(svp.get(0).stringValue()) - RU.FIRST_SLOT)
          * RU.DT_SLOT_SIZE  + RU.FIRST_SLOT;
            
        if ( f.deft().get(slot_desc_idx).type() == RU.MULTISLOT &&
            (svp.size() < 2 || svp.get(1).type() != RU.LIST))
          {            
            // wrap the new value in a list before adding
            ValueVector mf = new ValueVector();
            for (int j=1; j < svp.size(); j++)
              mf.add(svp.get(j));
            f.addValue(svp.get(0).stringValue(), new Value(mf, RU.LIST));
          }
        else 
          {
            // add the new value directly
            f.addValue(svp.get(0).stringValue(), svp.get(1));
          }
      }
        
    // get rid of the old one.
    context.engine().retract(fact.get(RU.ID).factIDValue());
        
    // get the new fact data
    fact = f.factData();
        
    // and assert the new fact
    return new Value(context.engine().assert(fact), RU.FACT_ID);
  }
}

// *** and *******************************************

class And implements Userfunction, ControlStructure
{
  public int skipFromHere() { return 0; }
  public String name() { return "and"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=1; i<vv.size(); i ++) 
      {
        Value v = vv.get(i);
        switch (v.type())
          {
          case RU.FUNCALL:
            v = Funcall.execute(context.expandAction(v.funcallValue()),
                                context);
            /* FALL THROUGH */

          default:
            if (v.equals(Funcall.s_false))
              return Funcall.s_false;
          }
      }
    return Funcall.s_true;
        
  }
}
// *** or *******************************************
class Or implements Userfunction, ControlStructure
{
  public int skipFromHere() { return 0; }
  public String name() { return "or"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=1; i<vv.size(); i ++) 
      {
        Value v = vv.get(i);
        switch (v.type())
          {
          case RU.FUNCALL:
            v = Funcall.execute(context.expandAction(v.funcallValue()),
                                context);
            /* FALL THROUGH */

          default:
            if (!v.equals(Funcall.s_false))
              return Funcall.s_true;
          }
      }
    return Funcall.s_false;
  }
}
        
// *** not *******************************************

class Not implements Userfunction
{
  public String name() { return "not"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    if (vv.get(1).equals(Funcall.s_false))
      return Funcall.s_true;
    else
      return Funcall.s_false;
  }
}
// *** eq *******************************************
 
class Eq implements Userfunction
{
  public String name() { return "eq"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=2; i<vv.size(); i ++) 
      {
        if (!vv.get(i).equals(vv.get(1)))
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}
// *** eq* *******************************************
 
class EqStar implements Userfunction
{
  public String name() { return "eq*"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=2; i<vv.size(); i ++) 
      {
        if (!vv.get(i).equalsStar(vv.get(1)))
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}

// *** = *******************************************


class Equals implements Userfunction
{
  public String name() { return "="; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=2; i<vv.size(); i ++) 
      {
        if (!(vv.get(i).numericValue() == vv.get(1).numericValue()))
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}
        
// *** <> *******************************************

class NotEquals implements Userfunction
{
  public String name() { return "<>"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=2; i<vv.size(); i ++) 
      {
        if (vv.get(i).numericValue() == vv.get(1).numericValue())
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}
        
// *** > *******************************************

class Gt implements Userfunction
{
  public String name() { return ">"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=1; i<vv.size()-1; i ++) 
      {
        double value1 = vv.get(i).numericValue();
        double value2 = vv.get(i+1).numericValue();
            
        if (!(value1 > value2))
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}
        
// *** < *******************************************

class Lt implements Userfunction
{
  public String name() { return "<"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=1; i<vv.size()-1; i ++) 
      {
        double value1 = vv.get(i).numericValue();
        double value2 = vv.get(i+1).numericValue();
            
        if (!(value1 < value2))
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}
        
// *** >= *******************************************

class GtOrEq implements Userfunction
{
  public String name() { return ">="; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=1; i<vv.size()-1; i ++) 
      {
        double value1 = vv.get(i).numericValue();
        double value2 = vv.get(i+1).numericValue();
            
        if (!(value1 >= value2))
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}
// *** <= *******************************************
class LtOrEq implements Userfunction
{
  public String name() { return "<="; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=1; i<vv.size()-1; i ++) 
      {
        double value1 = vv.get(i).numericValue();
        double value2 = vv.get(i+1).numericValue();
          
        if (!(value1 <= value2))
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}
// *** neq *******************************************

class Neq implements Userfunction
{
  public String name() { return "neq"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=2; i<vv.size(); i ++) 
      {
        if (vv.get(i).equals(vv.get(1)))
          return Funcall.s_false;
      }
    return Funcall.s_true;
  }
}
// *** mod *******************************************

class Mod extends Fastfunction
{
  public String name() { return "mod"; }
  public Value call(ValueVector vv, Context context, Value v)
       throws ReteException
  {
    int d1 = (int) vv.get(1).numericValue();
    int d2 = (int) vv.get(2).numericValue();
        
    return v.resetValue(d1 % d2, RU.INTEGER);
       
  }
}
// *** + *******************************************
class Plus extends Fastfunction
{
  public String name() { return "+"; }
  public Value call(ValueVector vv, Context context, Value v)
       throws ReteException
  {
    double sum = 0;
    int type = RU.INTEGER;
    int size = vv.size();
    for (int i=1; i<size; i++) 
      {
        Value arg = vv.get(i);
        sum += arg.numericValue();
        if (arg.type() == RU.FLOAT)
          type = RU.FLOAT;
          
      }

    return v.resetValue(sum, type);      
  } 
}
// *** * *******************************************

class Times extends Fastfunction
{
  public String name() { return "*"; }
  public Value call(ValueVector vv, Context context, Value v)
       throws ReteException
  {
    double product = 1;
    int type = RU.INTEGER;
    int size = vv.size();
    for (int i=1; i<size; i++) 
      {
        Value arg = vv.get(i);
        product *= arg.numericValue();
        if (arg.type() == RU.FLOAT)
          type = RU.FLOAT;
      }
                                  
    return v.resetValue(product, type);

  } 
}
// *** - *******************************************
class Minus extends Fastfunction
{
  public String name() { return "-"; }
  public Value call(ValueVector vv, Context context, Value v)
       throws ReteException
  {
    Value arg = vv.get(1);
    int type = arg.type();
    double diff = arg.numericValue();
    int size = vv.size();
    for (int i=2; i<size; i++) 
      {
        arg = vv.get(i);
        diff -= arg.numericValue();
        if (arg.type() == RU.FLOAT)
          type = RU.FLOAT;
      }

    return v.resetValue(diff, type);

  } 
}
// *** / *******************************************
class Divide extends Fastfunction
{
  public String name() { return "/"; }
  public Value call(ValueVector vv, Context context, Value v) 
       throws ReteException
  {
    double quotient = vv.get(1).numericValue();
    int size = vv.size();
    for (int i=2; i<size; i++) 
      {
        quotient /= vv.get(i).numericValue();
      }
    return v.resetValue(quotient, RU.FLOAT);


  } 
}

// *** sym-cat *******************************************
class SymCat extends Fastfunction
{
  public String name() { return "sym-cat"; }
  public Value call(ValueVector vv, Context context, Value v)
       throws ReteException
  {

    StringBuffer buf = new StringBuffer( "" );
    for ( int i = 1; i < vv.size( ); i++ )
      {
        Value val = vv.get(i);
        if (val.type() == RU.STRING)
          buf.append(val.stringValue());
        else
          buf.append (val.toString());
      }
    
    return v.resetValue( buf.toString( ), RU.ATOM );
  }
}

// *** facts *******************************************

class FactsRules implements Userfunction
{
  static final int FACTS=0, RULES=1;
  static final String[] s_names = {"facts", "rules"};
  private int m_name;
  FactsRules(int name) { m_name = name; }
  public String name() { return s_names[m_name]; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    switch(m_name)
      {
      case FACTS:
        context.engine().showFacts();
        break;
      case RULES:
        context.engine().showRules();
      }
    return Funcall.s_true;      
  } 
}

// *** store, fetch **************************************

class StoreFetch implements Userfunction
{
  static final int STORE=0, FETCH=1;
  static final String[] s_names = {"store", "fetch"};
  private int m_name;
  StoreFetch(int name) { m_name = name; }
  public String name() { return s_names[m_name]; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    Value v;
    switch(m_name)
      {
      case STORE:
        v = context.engine().store(vv.get(1).stringValue(),
                                   new Value(vv.get(2)));
        if (v != null)
          return v;
        else
          return Funcall.s_nil;
      case FETCH:
      default:
        v = context.engine().fetch(vv.get(1).stringValue());
        if (v != null)
          return v;
        else
          return Funcall.s_nil;
      }
  } 
}

// *** HaltEtc *******************************************
class HaltEtc implements Userfunction
{
  static final int HALT=0, EXIT=1, CLEAR=2, RUN=3, RESET=4, RETURN=5;
  static final String[] s_names = {"halt", "exit", "clear", "run",
                                   "reset", "return"};
  private int m_name;
  HaltEtc(int name) { m_name = name; }
  public String name() { return s_names[m_name]; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    switch (m_name)
      {
      case HALT:
        context.engine().halt();
        break;
      case EXIT:
        System.exit(0);
        break;
      case CLEAR:
        context.engine().clear();
        break;
      case RUN:
        if (vv.size() == 1)
          context.engine().run();
        else
          context.engine().run(vv.get(1).intValue());
        break;
      case RETURN:
        return context.setReturnValue(vv.get(1));        
      case RESET:
        context.engine().reset();
        break;
      }
    return Funcall.s_true;
  }
}

// *** watch *******************************************
class Watch implements Userfunction
{
  static final int WATCH=0, UNWATCH=1;
  static final String s_names[] = {"watch", "unwatch"};
  private int m_name;
  public String name() { return s_names[m_name]; }
  Watch(int name) { m_name = name; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    boolean cmd = (m_name == WATCH);
    String what = vv.get(1).stringValue();
    if (what.equals("rules"))
      context.engine().watchRules(cmd);
    else if (what.equals("facts"))
      context.engine().watchFacts(cmd);
    else if (what.equals("compilations"))
      context.engine().watchCompilations(cmd);
    else if (what.equals("activations"))
      context.engine().watchActivations(cmd);
    else if (what.equals("all")) 
      {
        context.engine().watchFacts(cmd);
        context.engine().watchRules(cmd);
        context.engine().watchCompilations(cmd);
        context.engine().watchActivations(cmd);
      }
    else
      throw new ReteException("Funcall::Execute", "watch: can't watch" ,
                              what);

    return Funcall.s_true;
  }
}

// *** jess versions  *******************************************

 
class JessVersion extends Fastfunction
{
  static final int NUMBER=0, STRING=1;
  static final String[] s_names={"jess-version-number",
                                 "jess-version-string"};
  private int m_name;
  JessVersion(int name) {m_name = name;}
  public String name() { return s_names[m_name]; }
  public Value call(ValueVector vv, Context context, Value v)
       throws ReteException
  {
    switch (m_name)
      {
      case NUMBER:
        return v.resetValue(4.4, RU.FLOAT);
      default:
        return v.resetValue("Jess Version 4.4 3/18/99", RU.STRING);
      }        
  }
}

// *** load-facts ***********************************************

class LoadFacts implements Userfunction
{
  public String name() { return "load-facts"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String s = "ERROR";
    InputStream f;
    if (context.engine().display().applet() == null) 
      {
        try 
          {
            f = new FileInputStream(vv.get(1).stringValue());
          }
        catch (IOException t) 
          {
            throw new ReteException("_load_facts::call",
                                    "I/O Exception",
                                    t.toString());
          }
        
      }
    else 
      {
        try 
          {
            URL url = new URL(context.engine().display().applet().getDocumentBase(),
                              vv.get(1).stringValue());          
            f = url.openStream();
          } 
        catch (Exception t) 
          {
            throw new ReteException("_load_facts::call",
                                    "Network error",
                                    t.toString());
          }
      }
      
    // OK, we have a stream. Now the tricky part!

    Jesp jesp = new Jesp(f, context.engine());

    return jesp.loadFacts();
  }
}

// *** save-facts ***********************************************


class SaveFacts implements Userfunction
{
  public String name() { return "save-facts"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    StringBuffer s = new StringBuffer();
    PrintStream f;
    if (context.engine().display().applet() == null) 
      {
        try 
          {
            f = new PrintStream(new FileOutputStream(vv.get(1).stringValue()));
          } 
        catch (IOException t) 
          {
            throw new ReteException("_save_facts::call", 
                                    "I/O Exception",
                                    t.toString());
          }
        
      }
    else 
      {
        try 
          {
            URL url = new URL(context.engine().display().applet().getDocumentBase(),
                              vv.get(1).stringValue());          
            URLConnection urlc  = url.openConnection();
            urlc.setDoOutput(true);
            f = new PrintStream(urlc.getOutputStream());
          }
        catch (Exception t) 
          {
            throw new ReteException("_load_facts::call",
                                    "Network error",
                                    t.toString());
          }
      }
      
    // OK, we have a stream. Now the tricky part!
    if (vv.size() > 2) 
      {
        for (int i=2; i< vv.size(); i++) 
          {
            s.append(context.engine().ppFacts(vv.get(i).stringValue()));
          
          }
      }
    else 
      s.append(context.engine().ppFacts());

    f.println(s.toString());
    f.close();
    return Funcall.s_true;
      
  } 
}

class AssertString extends Fastfunction
{
  public String name() { return "assert-string"; }
  public Value call(ValueVector vv, Context context, Value v)
       throws ReteException
  {
    String fact = vv.get(1).stringValue();
    return v.resetValue(context.engine().assertString(fact), RU.FACT_ID);      

  }
}

//***********************************************************//
// Karl Mueller NASA/GSFC Code 522.2 
// (Karl.R.Mueller@gsfc.nasa.gov)
// 26.January.1998
//
// *** retract-string ***************************************
// Added function to retract fact as a string
//***********************************************************//

class RetractString implements Userfunction
{
  public String name() { return "retract-string"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    for (int i=1; i< vv.size(); i++) 
      {
        context.engine().retractString(vv.get(i).stringValue());
      }
    return Funcall.s_true;
  }
}

class UnDefrule implements Userfunction
{
  public String name() { return "undefrule"; }
  public Value call(ValueVector vv, Context context) throws ReteException
  {
    String rulename = vv.get(1).stringValue();
    return context.engine().unDefrule(rulename);      

  } 

}    

// ************************************************************
// Fetch a variable by name. ReteCompiler uses this to fetch
// defglobals in rule LHSs.
// ************************************************************

class ExtractGlobal implements Userfunction
{
  public String name()
  {
    return "get-var";
  }

  public Value call(ValueVector vv, Context context)
       throws ReteException
  {
    return context.findBinding(vv.get(1).stringValue()).m_val;
  }
  
}


