/** **********************************************************************
 * Node containing an arbitrary list of tests; used for TEST CE's and also
 * the base class for join nodes.
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;

import java.util.*;

public class NodeTest extends Node
{
  
  /**
    The tests this node performs
    */

  Vector m_tests;
  Test[] m_localTests;

  /**
    Constructor
    */

  NodeTest(Rete engine) 
  {
    super(engine);
    m_tests = new Vector();
  }

  void complete() 
  {
    // freeze the tests vector
    m_localTests = new Test[m_tests.size()];
    for (int i=0; i< m_tests.size(); i++)
      m_localTests[i] = (Test) m_tests.elementAt(i); 
  }

  // Bare NodeTests can only have Test1's in them.
  void addTest(int test, int slot_idx, Value v) 
       throws ReteException
  {
    addTest(test, slot_idx, -1, v); 
  }

  void addTest(int test, int slot_idx, int slot_sidx, Value v) 
       throws ReteException
  {
    Test t;

    Funcall f = v.funcallValue();
    
    if (f.get(0).equals("java-test"))
      {
        String name = f.get(1).stringValue();
        try
          {
            t = (Test) Class.forName(name).newInstance();
          }
        catch (ClassCastException cce)
          {
            throw new ReteException("NodeTest.addTest",
                                    "Class for java-test must implement jess.Test:", name);
          }
        catch (ClassNotFoundException cnfe)
          {
            throw new ReteException("NodeTest.addTest",
                                    "Can't find class for java-test:", name);
          }
        catch (IllegalAccessException iae)
          {
            throw new ReteException("NodeTest.addTest",
                                    "Class for java-test is not public:",name);
          }
        catch (InstantiationException iae)
          {
            throw new ReteException("NodeTest.addTest",
                                    "Can't create java-test object: ",
                                    iae.getMessage());
          }
      }
            
    else
      t = new Test1(test, slot_idx, slot_sidx, v, null);
    
    m_tests.addElement(t);    
  }

  // The classic 'you should never inherit from concrete classes' problem!

  void addTest(int test, int token_idx, int left_idx, int right_idx) 
       throws ReteException
  {
    throw new ReteException("NodeTest:addtest",
                            "Can't add Test2s to this class", "");

  }

  void addTest(int test, int token_idx, int left_idx, int leftSub_idx,
               int right_idx, int rightSub_idx)
       throws ReteException
  {
    throw new ReteException("NodeTest:addtest",
                            "Can't add Test2s to this class", "");
  }

  /**
    For our purposes, two Node2's are equal if every test in
    one has an equivalent test in the other, and if the test vectors are
    the same size. The subclass NodeNot2 should never be shared, so we'll
    report unequal always. This routine is used during network compilation,
    not at runtime.
    */

  boolean equals(NodeTest n) 
  {
    if (this == n)
      return true;
      
    if (this.getClass() != n.getClass() ||
        this instanceof NodeNot2 ||
        n instanceof NodeNot2 ||
        n.m_tests.size() != m_tests.size())
      return false;

  outer_loop:
    for (int i=0; i<m_tests.size(); i++) 
      {
        // Test1 nodes hold function call tests. They're too
        // complicated to try to share.
        if (m_tests.elementAt(i) instanceof Test1)
          return false;
        Object t1 = m_tests.elementAt(i);
        for (int j=0; j<m_tests.size(); j++) {
          if (t1.equals(n.m_tests.elementAt(j)))
            continue outer_loop;
        }
        return false;
      }
    return true;
  }

  boolean callNode(Token token, int callType) throws ReteException
  {
    if (token.m_tag == RU.CLEAR)
      {
        passAlong(token);
        return true;
      }
    

    boolean result = runTests(token);
    if (result)
      {
        Successor [] sa = m_localSucc;
        int nsucc = sa.length;
        token = new Token(token, NodeNot2.s_nullFact);
        for (int j=0; j<nsucc; j++) 
          {
            Successor s = sa[j];
            s.m_node.callNode(token, s.m_callType);
          }
      }
    return result;
  }

  boolean runTests(Token token) throws ReteException
  {
    int ntests = m_localTests.length;

    int markF = m_cache.markFuncall();
    int markV = m_cache.markValue();
    
    try
      {
        for (int i=0; i < ntests; i++) 
          {
            if (!m_localTests[i].doTest(token, null, m_cache, m_engine))
              return false;
          }        
        return true;
      }
    catch (ReteException re)
      {
        re.addContext("'test' CE: " + toString());
        throw re;
      }
    finally
      {
        m_cache.restoreFuncall(markF);
        m_cache.restoreValue(markV);
      }
  }

  /**
    Describe myself
    */
  
  public String toString() 
  {
    StringBuffer sb = new StringBuffer(256);
    sb.append("[NodeTest ntests=");
    sb.append(m_tests.size());
    sb.append(" ");
    for (int i=0; i<m_tests.size(); i++)
      {
        sb.append(m_tests.elementAt(i).toString());
        sb.append(" ");
      }
    sb.append(";usecount = ");
    sb.append(m_usecount);
    sb.append("]");
    return sb.toString();
  }

}





