/** **********************************************************************
 * Rete machine: execute the built Network
 *
 * $Id$
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */

package jess;

import java.io.*;
import java.util.*;


public class Rete 
{

  /**
    Current conflict resolution strategy
   */
  private Strategy m_strategy;


  /**
    Context for executing global functions
    */

  private Context m_globalContext;
  public final Context globalContext() { return m_globalContext; }

  /**
    Where to draw myself
    */

  private ReteDisplay m_display;
  public final ReteDisplay display() { return m_display; }

  /**
    Whether to log stuff
    */


  private boolean m_watchRules;
  final public void watchRules(boolean val)
  { m_watchRules = val; }
  final public boolean watchRules()
  { return m_watchRules; }

  private boolean m_watchFacts;
  final public void watchFacts(boolean val)
  { m_watchFacts = val; }
  final public boolean watchFacts()
  { return m_watchFacts; }

  private boolean m_watchCompilations;
  final public void watchCompilations(boolean val)
  { m_watchCompilations = val; }
  final public boolean watchCompilations()
  { return m_watchCompilations; }

  private boolean m_watchActivations;
  final public void watchActivations(boolean val)
  { m_watchActivations = val; }
  final public boolean watchActivations()
  { return m_watchActivations; }

  private boolean m_resetGlobals = true;
  final public void setResetGlobals(boolean reset)
  { m_resetGlobals = reset; }
  final public boolean getResetGlobals()
  { return m_resetGlobals; }

  public static final int INSTALL=0, ACTIVATE=1, EVERY_TIME=2;
  private int m_evalSalience = INSTALL;
  final public void setEvalSalience(int method) throws ReteException
  {
    if (method < INSTALL || method > EVERY_TIME)
      throw new ReteException("Rete::setEvalSalience",
                              "Invalid value", String.valueOf(method));
    m_evalSalience = method;
  }
  final public int getEvalSalience()
  { return m_evalSalience; }
  

  /**
    Successively incremented ID for asserted facts.
    */
  private Object m_idLock = new Object();
  private int m_nextFactId;
  int nextFactId() { synchronized (m_idLock) {return m_nextFactId++;}}

  /**
    Successively incremented ID for new rules.
    */

  private int m_nextRuleId;
  int nextRuleId() { synchronized (m_idLock) {return m_nextRuleId++;}}

  /**
    Deftemplates are unique to each interpreter.
    */

  private Hashtable m_deftemplates = new Hashtable(101);

  /**
    Deffacts are unique to each interpreter.
    */

  private Hashtable m_deffacts = new Hashtable();

  /**
    Defglobals are unique to each interpreter.
    */

  private Hashtable m_defglobals = new Hashtable();

  private Hashtable m_functions = new Hashtable(101);

  /**
    Routers are kept in two hashtables: input ones and output ones.
    Names that are read-write are kept in both tables as separate entries.
    This means we don't need a special 'Router' class.

    Every input router is wrapped in a BufferedInputStream so we get reliable
    treatment of end-of-line. We need to keep track of the association, so
    we keep the original stream paired with the wrapper in m_inWrappers.

    Console-like streams act differently than file-like streams under
    read and readline , so when you cerate a router, you need to specify
    how it should act.
    */

  private Hashtable m_outRouters = new Hashtable(13);
  private Hashtable m_inRouters = new Hashtable(13);
  private Hashtable m_inWrappers = new Hashtable(13);
  private Hashtable m_inModes = new Hashtable(13);

  public void addInputRouter(String s, InputStream is, boolean consoleLike)
  { 
    synchronized (m_inRouters)
      {
        synchronized (m_inWrappers)
          {
            Tokenizer t = (Tokenizer) m_inWrappers.get(is);
            if (t == null)
              t = new Tokenizer(is);
            
            m_inRouters.put(s, is);
            m_inWrappers.put(is, t);
            m_inModes.put(s, new Boolean(consoleLike));
          }
      }
  }
  public void addOutputRouter(String s, OutputStream os)
  { m_outRouters.put(s, os); }

  public void removeInputRouter(String s)
  { m_inRouters.remove(s); }
  public void removeOutputRouter(String s)
  { m_outRouters.remove(s); }

  public InputStream getInputRouter(String s)
  { return (InputStream) m_inRouters.get(s);}

  public Tokenizer getInputWrapper(InputStream is)
  { return (Tokenizer) m_inWrappers.get(is);}

  public boolean getInputMode(String s)
  { return ((Boolean) m_inModes.get(s)).booleanValue();}

  public OutputStream getOutputRouter(String s)
  { return (OutputStream) m_outRouters.get(s);}
  
  /**
    The fact-list is unique to each interpreter.
    */
  
  private Vector m_facts = new Vector();

  /**
    The rule base is unique to each interpreter.
    */

  private Hashtable m_rules = new Hashtable();

  /**
    The agenda is unique to each interpreter.
    */

  private Vector m_activations = new Vector();

  /**
    Each interpreter has its own compiler object
    */

  private ReteCompiler m_compiler = new ReteCompiler(this);

  public synchronized final ReteCompiler compiler() { return m_compiler; }

  /**
    Flag for (halt) function
    */

  private boolean m_halt;

  /**
    Stuff to help us parse Jess code
    */

  private TextInputStream m_tis;
  private Jesp  m_jesp;

  /**
   * Stuff to help us communicate between Jess and Java
   * Things stored with store() are accessible from Jess with (fetch),
   * and vice-versa!
   */
  private Hashtable m_storage = new Hashtable();
  public Value store(String name, Value val)
  { return (Value) m_storage.put(name, val); }
  public Value store(String name, Object val)
  { 
    try
      {
        return (Value) m_storage.put(name,
                                     new Value(val, RU.EXTERNAL_ADDRESS));
      }
    catch (ReteException re) { return null; /* Can't happen */ }
  }
  public Value fetch(String name)
  { return (Value)  m_storage.get(name); }
  

  /**
    Constructor
    */

  public Rete(ReteDisplay display)
  {
    m_display = display;
    m_globalContext = new Context(this);

    // set up default routers
    addInputRouter("t", display.stdin(), true);
    addOutputRouter("t", display.stdout());
    addInputRouter("WSTDIN", getInputRouter("t"), true);
    addOutputRouter("WSTDOUT", getOutputRouter("t"));
    addOutputRouter("WSTDERR", getOutputRouter("t"));

    m_tis = new TextInputStream(true);
    m_jesp = new Jesp(m_tis, this);

    try
      {
        m_strategy = (Strategy) Class.forName("jess.depth").newInstance();
      }
    catch (Throwable t)
      {
        display.stdout().println("Rete::Rete: Can't load default strategy");
        display.stdout().println("class jess.depth. ");
        display.stdout().println(t);
        System.exit(-1);
      }    

    try
      {
        Funcall.loadIntrinsics(this);
      }
    catch (Throwable t)
      {
        display.stdout().println("Rete::Rete: Can't load built-in functions.");
        display.stdout().println("Exiting...");
        display.stdout().println(t);
        System.exit(-1);
      }    
  }

  public PrintStream errStream()
  {
    synchronized (m_outRouters)
      {
        // Coerce to PrintStream;
        PrintStream ps;
        OutputStream os = getOutputRouter("WSTDERR");
        if (os instanceof PrintStream)
          ps = (PrintStream) os;
        else
          {
            ps = new PrintStream(os);
            addOutputRouter("WSTDERR", ps);
          }
        return ps;
      }
  }

  public PrintStream outStream()
  {
    synchronized (m_outRouters)
      {
        // Coerce to PrintStream;
        PrintStream ps;
        OutputStream os = getOutputRouter("WSTDOUT");
        if (os instanceof PrintStream)
          ps = (PrintStream) os;
        else
          {
            ps = new PrintStream(os);
            addOutputRouter("WSTDOUT", ps);
          }
        return ps;
      }
  }

  private Vector m_clearables = new Vector();
  public void addClearable(Clearable c)
  { m_clearables.addElement(c); }
  public void removeClearable(Clearable c)
  { m_clearables.removeElement(c); }

  /*
    Reinitialize engine
    Thanks to Karl Mueller for idea
    */

  public synchronized void clear() throws ReteException
  {
    watchRules(false);
    watchFacts(false);
    watchCompilations(false);
    watchActivations(false);
    m_halt = false;
    m_nextFactId = m_nextRuleId = 0;
    m_deftemplates.clear();
    m_deffacts.clear();
    m_defglobals.clear();
    m_storage.clear();

    Enumeration e = m_clearables.elements();
    while (e.hasMoreElements())
      ((Clearable) e.nextElement()).clear();

    m_facts.removeAllElements();
    m_compiler = new ReteCompiler(this);
    if (m_rules.size() != 0)
      m_display.addDefrule((Defrule) m_rules.elements().nextElement());
    m_rules.clear();
    m_activations.removeAllElements();

    // Undefine Deffunctions, but nothing else
    Hashtable temp = new Hashtable(100);

    for (Enumeration ee = m_functions.keys(); ee.hasMoreElements();)
      {
        Object o = ee.nextElement();
        Userfunction uf = (Userfunction) m_functions.get(o);
        if (! (uf instanceof Deffunction))
          temp.put(o, uf);
      }
    m_functions=temp;   
    
    m_resetables.removeAllElements();
    m_clearables.removeAllElements();

    System.gc();
  }

  private Vector m_resetables = new Vector();
  public void addResetable(Resetable c)
  { m_resetables.addElement(c); }
  public void removeResetable(Clearable c)
  { m_resetables.removeElement(c); }

  /**
    Reset the interpreter. Remove all facts, flush the network,
    remove all activations.
    */

  void removeFacts() throws ReteException
  {
    synchronized (m_compiler)
      {
        // remove all existing facts
        // This Token tag is a special command. All 1-input nodes
        // just pass it along; all two-input nodes clear both memories.
        
        Fact nf = new Fact("*CLEAR*", RU.ORDERED_FACT, this);
        ValueVector vv = nf.factData();
        Token t = new Token(RU.CLEAR, vv);
        processToken(t);
        System.gc();
        m_facts = new Vector();
        m_display.retractFact(vv);
      }
  }

  public void reset() throws ReteException 
  {
    
    synchronized (m_compiler)
      {
        removeFacts();
        m_nextFactId = 0;
        
        // assert the initial-fact
        Fact init = new Fact("initial-fact", RU.ORDERED_FACT, this);
        assert(init.factData());
        
        // Defglobals, Deffacts get taken care of in here!
        Enumeration e = m_resetables.elements();
        while (e.hasMoreElements())
          ((Resetable) e.nextElement()).reset();
      }
  }

  /**
    Assert a fact, as a String
    */

  public int assertString(String s) throws ReteException 
  {
    StringBufferInputStream sbis;
    try
      {
        synchronized (m_tis)
          {
            m_tis.clear();
            m_jesp.clear();
            m_tis.appendText(s);
            Fact f = m_jesp.parseFact();
            return assert(f.factData());
          }
      }
    catch (Exception t)
      {
        throw new ReteException("Rete::AssertString", t.toString(), s);
      }
  }

  /**
    Assert a fact, as a ValueVector
    */

  public int assert(ValueVector f) throws ReteException 
  {
    synchronized (m_compiler)
      {
        // First, expand any multifields if ordered fact
        if ( f.get(RU.DESC).descriptorValue() == RU.ORDERED_FACT)
          for (int i=0; i<f.size(); i++)
            if (f.get(i).type() == RU.LIST) 
              {
                ValueVector nf = new ValueVector(f.size());
                for (int j=0; j<f.size(); j++) 
                  {
                    Value v = f.get(j);
                    if (v.type() != RU.LIST)
                      nf.add(v);
                    else 
                      {
                        ValueVector vv = v.listValue();
                        for (int k=0; k < vv.size(); k++) 
                          {
                            nf.add(vv.get(k));
                          }
                      }
                  }
                f = nf;
                break;
              }
        
        // find any old copy
        ValueVector of = findFact(f);
        
        if (of != null)
          return -1;
        
        // insert the new fact
        f.set(new Value(m_nextFactId++, RU.FACT_ID), RU.ID);
        
        m_facts.addElement(f);
        Token t = new Token(RU.ADD, f);
        processToken(t);
        
        m_display.assertFact(f);
        
        if (m_watchFacts)
          outStream().println(" ==> "  +
                              new Fact(f,this).toString());

        return f.get(RU.ID).factIDValue();
      }
  }

  /** ***********************************************************
   * Karl Mueller NASA/GSFC Code 522.2 
   * (Karl.R.Mueller@gsfc.nasa.gov)
   * 27.January.1998
   *
   * Retract a fact as a string
   * ************************************************************/

  public void retractString(String s) throws ReteException 
  {
    try
      {
        synchronized (m_tis)
          {
            m_tis.clear();
            m_jesp.clear();
            m_tis.appendText(s);
            Fact f = m_jesp.parseFact();
            retract(f.factData());
          }
      }
    catch (Exception t)
      {
        throw new ReteException("Rete::retractString", t.toString(), s);
      }
  }
  

  /**
    Retract a fact, as a ValueVector
    */

  public void retract(ValueVector f) throws ReteException 
  {
    synchronized (m_compiler)
      {
        ValueVector found;
        if ((found = findFact(f)) != null) 
          _retract(found);
      }
  }
  
  /**
    Retract a fact by ID, used by rule RHSs.
    */

  public void retract(int id) throws ReteException 
  {
    synchronized (m_compiler)
      {
        ValueVector found;
        if ((found = findFactByID(id)) != null)
          _retract(found);
      }
  }

  /*
    Do the actual retracting - f MUST be a valid fact
    */

  void _retract(ValueVector f) throws ReteException
  {
    synchronized (m_compiler)
      {
        m_facts.removeElement(f);
        
        Token t = new Token(RU.REMOVE, f);
        processToken(t);
        
        if (m_watchFacts)
          outStream().println(" <== "  +
                              " " + new Fact(f,this).toString());
        m_display.retractFact(f);      

      }    
  }

  /**
    This 'find' is used by the retract that rules use. Consider return value
    to be READ-ONLY!
    */

  public ValueVector findFactByID(int id) throws ReteException 
  {
    for (Enumeration e = m_facts.elements(); e.hasMoreElements();) 
      {
        ValueVector tf = (ValueVector) e.nextElement();
        if (tf.get(RU.ID).factIDValue() == id)
          return tf;
      }
    return null;
  }

  /**
    Does a given fact (as a ValueVector) exist? (We're looking for identical
    data, but the ID can differ)
    */

  private ValueVector findFact(ValueVector f) throws ReteException 
  {
    int size = m_facts.size();
    int fsize = f.size();
  outer_loop:
    for (int i=0; i < size; i++) 
      {
        ValueVector tf = (ValueVector) m_facts.elementAt(i);
        if (fsize != tf.size())
          continue;
        if (!f.get(RU.CLASS).equals(tf.get(RU.CLASS)))
          continue;
        for (int j=RU.FIRST_SLOT; j < fsize; j++) 
          {
            if (!f.get(j).equals(tf.get(j)))
              continue outer_loop;
          }
        return tf;
      }
    return null;
  }

  /**
    What the (rules) command calls
    */

  public void showRules() 
  {

    for (Enumeration e = m_rules.elements(); e.hasMoreElements();)
      {
        Defrule dr = (Defrule) e.nextElement();
        outStream().println(dr.toString());
      }
    outStream().println("For a total of " + m_rules.size() + " rules.");
  }
  
  /** Return the pretty print forms of all facts, as a big string */

  public String ppFacts(String name) 
  {
    StringBuffer sb = new StringBuffer(100);
    Fact fact = null;
    for (int i=0; i<m_facts.size(); i++) 
      {
        ValueVector f = (ValueVector) m_facts.elementAt(i);
        try 
          {
            if (!f.get(RU.CLASS).stringValue().equals(name)) continue;
            fact = new Fact(f, this);
          }
        catch (ReteException re) 
          {
            continue;
          }
        sb.append(fact);
        sb.append("\n");
      }
    return sb.toString();
  }

  public String ppFacts() throws ReteException
  {
    StringBuffer sb = new StringBuffer(1024);
    Fact fact = null;
    for (Enumeration e = m_facts.elements(); e.hasMoreElements();)
      {
        ValueVector f = (ValueVector) e.nextElement();
        fact = new Fact(f, this);
        sb.append(fact);
        sb.append("\n");
      }
    return sb.toString();
  }


  /**
    What the (facts) command calls
    */

  public void showFacts() throws ReteException 
  {
    for (Enumeration e = m_facts.elements(); e.hasMoreElements();)
      {
        ValueVector f = (ValueVector) e.nextElement();
        Fact fact = new Fact(f, this);
        outStream().print("f-");
        outStream().print(f.get(RU.ID).factIDValue() + "   ");
        outStream().println(fact.toString());
      }
    outStream().println("For a total of " + m_facts.size() + " facts.");
  }
  
  /** 
    These methods are to be used primarily for debugging purposes.
    Don't get too chummy with the way they work - they will change without
    notice.   
    */

  public Enumeration listDeffacts()
  { return m_deffacts.elements(); }

  public Enumeration listDeftemplates()
  { return m_deftemplates.elements(); }

  public Enumeration listDefrules()
  { return m_rules.elements(); }

  public Enumeration listFacts()
  { return m_facts.elements(); }

  public Enumeration listActivations()
  { return m_activations.elements(); }

  public Enumeration listDefglobals()
  { return m_defglobals.elements(); }

  public Enumeration listFunctions()
  { return m_functions.elements(); }
  
  /**
    Process a Token which represents a fact being added or removed.
    Eventually we should set this up so that the token gets dispatched
    to only one root of the pattern net - right now it presented to all of
    them!
    
    */

  private boolean processTokenOneNode(Token t, Node n) throws ReteException
  {
    synchronized (m_compiler)
      {
        return n.callNode(t, Node.SINGLE);
      }
  }
       
  private boolean processToken(Token t) throws ReteException 
  {
    synchronized (m_compiler)
      {
        Vector v = m_compiler.roots();
        
        // make sure the network is optimized
        m_compiler.freeze();
        
        for (Enumeration e = v.elements(); e.hasMoreElements();)
          {
            Node1 n = (Node1) ((Successor) e.nextElement()).m_node;
            if (processTokenOneNode(t, n))
              return true;
          }
        return false;
      }
  }

  /**
    Present all the facts on the agenda to a single rule.
    */

  void updateNodes(Hashtable n) throws ReteException 
  {    
    m_compiler.freeze();
    for (Enumeration e = m_facts.elements(); e.hasMoreElements();)
      {
        Token t = new Token(RU.UPDATE, (ValueVector) e.nextElement());
        for (Enumeration nodes = n.elements(); nodes.hasMoreElements();)
          processTokenOneNode(t, (Node) nodes.nextElement());
      }
  }

  /**
    Find a defrule object with a certain name
    */

  public final Defrule findDefrule(String name) 
  {
    return (Defrule) m_rules.get(name);
  }

  /**
    Find a deftemplate object with a certain name
    */

  Deftemplate findDeftemplate(String name) 
  {
    return ((Deftemplate) m_deftemplates.get(name));
  }

  /**
    find the deftemplate, if there is one, or create implied dt.
    */

  Deftemplate createDeftemplate(String name, int ordered)
       throws ReteException 
  {

    Deftemplate deft = findDeftemplate(name);
    if (deft != null) 
      {
        if (ordered != deft.m_ordered) 
          // looks like there are semantic errors in input.
          throw new ReteException("Rete::createDeftemplate",
                                  "Attempt to duplicate implied deftemplate:",
                                  name);
      }
    else  
      {
        // this is OK. Create an implied deftemplate if this is an
        // ordered pattern.
        if (ordered == RU.UNORDERED_FACT) 
          throw new ReteException("Pattern::FindDeftemplate",
                                  "Can't create implied unordered deftemplate",
                                  name);
        else
          deft = addDeftemplate(new Deftemplate(name,ordered));
        
      }
    return deft;
  }



  /**
    Creates a new deftemplate in this object. 
    Ensure that every deftemplate has a unique class name; silently
    ignore redefinitions!
    */

  public Deftemplate addDeftemplate(Deftemplate dt)
       throws ReteException 
  {
    synchronized (m_deftemplates)
      {
        String name = dt.name();
        if (m_deftemplates.get(name) == null) 
          {
            m_deftemplates.put(name, dt);
            m_display.addDeftemplate(dt);
          }
        return dt;
      }
  }

  /**
    Creates a new deffacts in this object
    */

  public Deffacts addDeffacts(Deffacts df) throws ReteException 
  {    
    m_deffacts.put(df.name(), df);
    m_display.addDeffacts(df);
    addResetable(df);
    return df;
  }

  /**
    Creates a new Defglobal in this object. Trick it into resetting 
    right now, regardless of the setting of resetGlobals.
    */

  public Defglobal addDefglobal(Defglobal dg) throws ReteException 
  {
    m_defglobals.put(dg.name(), dg);
    
    boolean oldReset = m_resetGlobals;
    try
      {
        m_resetGlobals = true;
        dg.reset();
      }
    finally
      {
        m_resetGlobals = oldReset;
      }

    addResetable(dg);
    return dg;
  }

  /**
    Creates a new function in this object
    Will happily destroy an old one.
    */

  public Userfunction addUserfunction(Userfunction uf) 
  {
    m_functions.put(uf.name(), uf);
    return uf;
  }

  /**
    A package simply calls AddUserfunction lots of times.
    */

  public Userpackage addUserpackage(Userpackage up) 
  {
    up.add(this);
    return up;
  }

  /**
    Find a userfunction, if there is one.
    */

  public Userfunction findUserfunction(String name) 
  {
    return (Userfunction) m_functions.get(name);
  }


  /**
    Creates a new defrule in this object
    */

  public final Defrule addDefrule(Defrule dr) throws ReteException 
  {
    synchronized (m_compiler)
      {
        unDefrule(dr.name());
        
        m_compiler.addRule(dr);
        m_rules.put(dr.name(), dr);
        m_display.addDefrule(dr);
        return dr;
      }
  }

  public final Value unDefrule(String name) throws ReteException 
  {
    synchronized (m_compiler)
      {
        Defrule odr = findDefrule(name);
        if (odr != null)
          {
            odr.remove(m_compiler.roots());
            m_rules.remove(name);
            for (Enumeration e=m_activations.elements(); e.hasMoreElements();)
              {
                Activation a = (Activation) e.nextElement();
                if (a.m_rule == odr)
                  standDown(a);
              }
            
            m_display.addDefrule(odr);
            return Funcall.TRUE();
          }
      }

    return Funcall.FALSE();
  }

  /**
    Info about a rule to fire.
    */

  void addActivation(Activation a) throws ReteException 
  {
    m_display.activateRule(a.m_rule);
    

    if (m_evalSalience != Rete.INSTALL)
      a.m_rule.evalSalience();

    m_strategy.addActivation(a, m_activations);
    
    if (m_watchActivations)
      outStream().println("==> Activation: " +
                          a.m_rule.name() + " : " +
                          factList(a.m_token));
  }

  public String setStrategy(Strategy s) throws ReteException
  {
    synchronized (m_activations)
      {
        String rv = m_strategy.name();
        m_strategy = s;
        Vector v = m_activations;
        m_activations = new Vector();
        for (Enumeration e = v.elements(); e.hasMoreElements();)
          addActivation((Activation) e.nextElement());
        return rv;
      }
  }


  /**
    An activation has been cancelled; forget it
    */

  void standDown(Activation a) throws ReteException 
  {
    ruleFired(a);
    if (m_watchActivations)
      outStream().println("<== Activation: " +
                                 a.m_rule.name() + " : " +
                                 factList(a.m_token));
  }

  /**
    An activation has been fired; forget it
    */

  void ruleFired(Activation a) 
  {
    m_display.deactivateRule(a.m_rule);
    m_activations.removeElement(a);
  }

  /**
    Return a string describing a list of facts
    */
  
  static String factList(Token t) throws ReteException 
  {    
    StringBuffer sb = new StringBuffer(100);
    boolean first = true;
    for (int i=0; i<t.size(); i++) 
      {
        if (!first)
          sb.append(",");
        int id = t.fact(i).get(RU.ID).factIDValue();
        if (id != -1)
          {
            sb.append(" f-");
            sb.append(id);
          }
        first = false;
      }
    return sb.toString();
  }

  /**
    Run the actual engine.
    */

  public int run() throws ReteException
  {
    int i=0, j;
    do
      {
        j = run(Integer.MAX_VALUE);
        i += j;
      }
    while (j > 0 && !m_halt);
    return i;
  }

  public synchronized int run(int max) throws ReteException 
  {
    int n = 0;
    int size = 0;
    m_halt = false;
    
    while (m_activations.size() > 0 && !m_halt && n < max) 
      {        
        Thread.yield();
        
        if (m_activations.size() > 0)
          {
            Activation a = (Activation) m_activations.elementAt(0);
            m_display.fireRule(a.m_rule);
            a.fire();
            ++n;
          }

        if (m_evalSalience == EVERY_TIME)
          {
            synchronized (m_activations)
              {
                Vector v = m_activations;
                m_activations = new Vector();

                for (Enumeration e = v.elements(); e.hasMoreElements();)
                  addActivation((Activation) e.nextElement());
              }
          }


      }
    return n;
  }
  
  /**
    Stuff to let Java code call functions inside of us.
    */
  
  
  public Value executeCommand(String cmd) throws ReteException 
  {
    synchronized (m_tis)
      {
        m_tis.clear();
        m_jesp.clear();
        m_tis.appendText(cmd);
        return m_jesp.parse(false);
      }
  }

  /**
    Jane, stop this crazy thing!
    */

  public void halt() 
  {
    m_halt = true;
  }

}

